package MIB::Quota;

use warnings;
use strict;
use utf8;
use feature "switch";
no warnings 'experimental::smartmatch';

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(iterate);

use Quota;
use Net::LDAP;
use YAML::XS qw(LoadFile);
use Number::Format qw/:subs/;
use Config::Find;


our $VERSION = '1.40';


my $conffile = Config::Find->find(name => 'ldapquota') or die "Can't find configuration file\n";
my $config = LoadFile($conffile);


my $ldap = Net::LDAP->new($config->{'ldap'}->{'host'}) or die "$@";
my $mesg = $ldap->bind;
END { $ldap->unbind() }
$mesg->code && die $mesg->error;

$mesg = $ldap->search(
	base => $config->{'ldap'}->{'base'},
	filter => $config->{'ldap'}->{'filter'},
);
$mesg->code && die $mesg->error;


my %defaults;
my @paths = keys %{$config->{'paths'}};

foreach my $path (@paths) {
	# allow recursive defaults even though the order of hash keys doesn't match the order of lines in the config file.
	# use sections "define1" through "define9" (and process them in that order and before "default") to hack around that.
	foreach my $section (map("define$_", (1..9)), "default") {
		while (my ($class, $ref) = each %{$config->{paths}->{$path}->{$section}}) {
			$defaults{$path}->{$class} = get_values($path, $section, $ref);
		}
	}
}


sub get_values {
	my ($path, $msg, $val) = @_;

	my $arr;
	given (ref($val)) {
		when ("") {
			given ($val) {
				when ("IGNORE") {
					return undef;
				};
				when ("NONE") {
					$arr = [ "NONE", "NONE", "NONE", "NONE" ];
				};
				when (/UNLIM(ITED)?/) {
					$arr = [ "UNLIM", "UNLIM", "UNLIM", "UNLIM" ];
				};
				default {
					my $def = $defaults{$path}->{$val};
					die "unknown value $val for $path ($msg)\n" unless $def;
					# Werte rauskopieren, damit nicht im Endeffekt eine Referenz in %defaults an iterate() übergeben wird.
					return {%{$def}};
				};
			};
		};
		when ("ARRAY") {
			$arr = $val;
		};
		default {
			die "unexpected reference ($_) in value $val for path $path ($msg)\n"
		};
	}

	my %values = ( kbyte_soft => 0, kbyte_hard => 1, inode_soft => 2, inode_hard => 3 ); # values are indices into $arr->[]

	# convert to kbytes/numbers
	# uc() as Number::Format only knows about K, M, G (and not k, m, g)
	foreach my $var ($values{kbyte_soft}, $values{kbyte_hard}) {
		given ($arr->[$var]) {
			when ("IGNORE") { die "cannot partially IGNORE an entry\n" };
			when (/^(UNLIM(ITED)?|0)$/) { $var = 0 };
			when (/^(NONE|1)$/) { $var = 1 };
			when (/^ *[^\d]/) { warn "entry $_ is not numeric\n"; continue };
			when (/\d *$/) { warn "entry $_ has no unit\n"; continue };
			default {
				my $x = unformat_number(uc($_)) or die "cannot unformat_number $_\n";
				$var = int(($x + 1023) / 1024);
			};
		}
	}
	foreach my $var ($values{inode_soft}, $values{inode_hard}) {
		given ($arr->[$var]) {
			when ("IGNORE") { die "cannot partially IGNORE an entry\n" };
			when (/^(UNLIM(ITED)?|0)$/) { $var = 0 };
			when (/^(NONE|1)$/) { $var = 1 };
			when (/^ *[^\d]/) { warn "entry $_ is not numeric\n"; continue };
			default {
				my $x = unformat_number(uc($_), base => 1000) or die "cannot unformat_number $_\n";
				$var = int($x);
			};
		}
	}

	return \%values;
}

sub get_usage {
	my ($arr) = @_;

	my %values = (
		kbyte_curr => 0, kbyte_soft => 1, kbyte_hard => 2, kbyte_timelimit => 3,
		inode_curr => 4, inode_soft => 5, inode_hard => 6, inode_timelimit => 7,
	); # values are indices into Quota::query output
	# Kilobytes
	$_ = $arr->[$_] // 0 foreach ($values{kbyte_curr}, $values{kbyte_soft}, $values{kbyte_hard});
	# Inodes
	$_ = $arr->[$_] // 0 foreach ($values{inode_curr}, $values{inode_soft}, $values{inode_hard});
	# Times
	$_ = $arr->[$_] // 0 foreach ($values{kbyte_timelimit}, $values{inode_timelimit});

	return \%values;
}

sub percent {
	my ($p, $q) = @_;
	return 0 unless $q; # oder lieber undef zurückgeben?
	return int(100 * $p / $q);
}

sub iterate {
	my ($function, $users) = @_;

	my ($all_users, %only_users);
	if (defined $users and @$users) {
		$only_users{$_} = 1 foreach (@$users);
	} else {
		$all_users = 1;
	}

	foreach my $entry ($mesg->entries) {
		my $uid   = $entry->get_value('uidnumber');
		die "No uidNumber attribute\n" unless $uid;
		my $user  = $entry->get_value('uid');
		die "No uid attribute\n" unless $user;

		next unless $all_users or $only_users{$user};

		my $class = $entry->get_value($config->{'ldap'}->{'attr'});
		if (defined $class) {
			$class = lc($class);
			next if $class eq 'ignore';
		} else {
			$class = "UNDEF";
		}

		foreach my $path (@paths) {
			my %values;

			my $name = $config->{paths}->{$path}->{name};
			unless (defined $name) { # use last pathname component instead
				$path =~ m{/([^/]+)$}; $name = $1;
			}

			if (defined $defaults{$path}->{$class}) {
				# Werte aus %defaults{$path}{$class} rauskopieren, damit nicht implizit %defaults geändert wird.
				# Die Referenz wird auf $values{target} übertragen und deren Werte können von der aufgerufenen Funktion verändert werden.
				# Es reicht ein shallow copy, weil es nur noch eine Ebene (kbyte/inodes_hard/soft) darunter gibt.
				$values{default} = {%{$defaults{$path}->{$class}}};
			} elsif (exists $defaults{$path}->{$class}) {
				$values{default} = undef;
			} else {
				warn "No $path quota for class $class\n";
			}
			if (exists $config->{paths}->{$path}->{$user}) {
				my $exception = $config->{paths}->{$path}->{$user};
				if ($exception =~ /^\d$/) {
					my $class_exception = $class . '_' . $exception;
					if (exists($defaults{$path}->{$class_exception})) {
						$exception = $class_exception;
					} else {
						warn "user $user: no level $exception in class $class for $path, ignoring exception\n";
						$exception = "";
					}
				}
				if ($exception) {
					$values{exception} = get_values($path, "user $user", $exception);
					$values{target} = $values{exception};
				} else {
					$values{target} = $values{default};
				}
			} else {
				$values{target} = $values{default};
			}
			unless (defined $values{target}) {
				# either neither class nor user entry (a reason to warn)
				# or an explicit IGNORE entry for class or user (no reason to warn)
				warn "No $path quota for user $user\n" unless (exists $values{default} or exists $values{exception});
				next;
			}

			## This can't be right ... normally getqcard is
			## used the get the canonical device for a path,
			## but this does not seem to work for nfs mounted
			## filesystems. So we check here if the path looks
			## like foo:/bar

			my $dev;
			if ( $path =~ m{^[^/]+:} ) {
				$dev = $path;
			} else {
				$dev = Quota::getqcarg($path);
			}
			my @query = Quota::query($dev, $uid);
			$values{error} = Quota::strerr() unless defined $query[0];
			$values{real} = get_usage(\@query);

			$values{real}->{kbyte_percent} = percent($values{real}->{kbyte_curr}, $values{real}->{kbyte_soft});
			$values{real}->{inode_percent} = percent($values{real}->{inode_curr}, $values{real}->{inode_soft});

			$values{$_} = eval("\$$_") foreach qw/user uid class dev path name config/;
				
			$function->({%values});

		}
	}
}
=pod

=head1 NAME

MIB::Quota - Interface between Quota and LDAP

=head1 SYNOPIS

=head1 DESCRIPTION

=head1 SEE ALSO

L<Quota>, L<Net::LDAP>, L<YAML>.

=cut

1;


